/*
	Copyright 1992 Chris Lewis
		All Rights Reserved

    See the LICENSE file for a full description of restrictions under which
    this software is provided.

    Function:		LaserJet driver, drawing code.
			Needs LJII (actually PCL4) or later (rules)
    
    Currently supports:
	line thickness set
	fill set
	circle & ellipse, c&e fill
	lines
	polygons
	splines
    
    TODO: 
	- polyfill.  Polyfill probably never unless somebody
	  gives me a polygon fill algorithm (hint hint!)
	- optimize emitrast some more.  (wastes up to 25 bytes per subsequent
	  raster line where the start point is within 2/3 inch.)

*/

#include "defs.h"

#if	defined(LJ) && defined(INSPECIAL)
#include "lj.h"
#include "math.h"

#ifndef	lint
static char SCCSid[] =
    "@(#)ljdraw.c: 1.7 Copyright 93/01/17 22:36:34 Chris Lewis";
#endif

static int ljLine(), ljPos(), lineseg();
int asetbit();

/*	Do ditroff drawing command in s from origX,origY.
	If opcode 0 just moveto
 */

int curthick = 1;	/* half of current thickness */
int curfill = 0;	/* solid black 1000 is solid white */

int drawDot();
#define sqr(x)	((x) * (x))

static long	rastcount, total;

#ifndef M_PI
#define	M_PI	3.141529
#endif
#define	RAD_C	(M_PI*2)

static int numx, numy;

ljDraw(origX, origY, opcode, ctindex, numbers, saves)
int origX, origY;
int opcode;
char *saves;
int ctindex;
short *numbers; {

    int i;

    /* map absolute beginning position. */
    origX = TROFF2LJX(origX);
    origY = TROFF2LJY(origY);

    if (!numx) {
	numx = pagewidth * LJRESOLUTION/TROFFRESOLUTION;
	numy = pagelength * LJRESOLUTION/TROFFRESOLUTION;
    }


    if (!opcode) {
	ljPos(origX, origY);
	return;
    }

    switch (opcode) {
	case 'f':
	    if (ctindex == 2)
		ctindex = 1;
	    if (!require(1, ctindex, "set fill", saves)) {
		curfill = numbers[0];
		if (curfill < 0)
		    curfill = 0;
		if (curfill > 1000)
		    curfill = 1000;
		DBP((D_PIC, "FILL: %d\n", curfill));
	    }
	    return;
	case 't':
	    if (ctindex == 2)
		ctindex = 1;
	    if (!require(1, ctindex, "set thick", saves)) {
		curthick = numbers[0] / 2;
		if (curthick < 1)
		    curthick = 1;
		DBP((D_PIC, "THICK: %d\n", curthick));
	    }
	    return;
    }

    DBP((D_CAT, "ljDraw: (%d,%d): %s\n", origX, origY, saves));

    switch (opcode) {
	case 'C': case 'c':
	    if (ctindex == 2)
		ctindex = 1;
    }

    for (i = 0; i < ctindex; i++)
	numbers[i] = numbers[i] * LJRESOLUTION / TROFFRESOLUTION;

    switch(opcode) {
	case 'l':
	    if (require(2, ctindex, "line", saves))
		return;
	    ljLine(origX, origY, origX + numbers[0], origY + numbers[1]);
	    return;

	case 'C':
	case 'c':
	    if (require(1, ctindex, "circle", saves))
		return;
	    numbers[1] = numbers[0];
	    ljCircle(origX + numbers[0] / 2, origY, numbers[0] / 2,
		drawDot, opcode == 'C' ? curfill: -1);
	    return;

	case 'E':
	case 'e':
	    if (require(2, ctindex, "ellipse", saves))
		return;
	    ljEllipse(origX + numbers[0] / 2, origY, numbers[0]/2, numbers[1]/2,
		opcode == 'E' ? curfill: -1);
	    return;

	case 'P':
	case 'p':
	    if (ctindex&1) {
		fprintf(stderr, "%s: need even number of points for polygon: %s\n",
		    progname, saves);
		return;
	    }
	    ljPoly(origX, origY, ctindex, numbers);
	    return;

	case 'a':
	    if (require(4, ctindex, "arc", saves))
		return;
	    ljArc(origX, origY, numbers[0], numbers[1], numbers[2], numbers[3]);
	    return;

	case '~': 
	    if (ctindex < 4 || (ctindex&1)) {
		fprintf(stderr, "%s: too few points to spline, or odd point count: %s\n",
		    progname, saves);
		return;
	    }
	    ljSpline(origX, origY, ctindex, numbers);
	    return;
	    
	default:
	    fprintf(stderr, "%s: invalid draw code %c (%s)\n", progname, opcode,
		saves);
	    return;
    }
}

/* Can't do polygon fill.  Probably never will - needs far more
   advanced techniques than I'm willing to do */

ljPoly(curx, cury, ctindex, numbers)
int curx, cury, ctindex;
short *numbers; {
    int i;

    DBP((D_PIC, "ljPoly\n"));

    for (i = 0; i < ctindex; i += 2) {
	ljLine(curx, cury, curx + numbers[i], cury + numbers[i+1]);
	curx += numbers[i];
	cury += numbers[i+1];
    }

}

int lx, ly;

ljSpline(ox, oy, ctindex, numbers)
int ox, oy, ctindex;
short *numbers; {

    int dx = numbers[0], dy = numbers[1];
    int sx = ox, sy = oy, tx = sx + dx, ty = sy + dy;
    int i;

    DBP((D_PIC, "ljSpline\n"));

    ljLine(sx, sy, lx = (sx + tx) / 2, ly = (sy + ty) / 2);

    for (i = 2; i < ctindex; i += 2) {
	int ux = ox + (dx += numbers[i]);
	int uy = oy + (dy += numbers[i+1]);

	flatten((sx + tx*5)/6, (sy + ty*5)/6, (tx*5 + ux) / 6,
		(ty*5 + uy) / 6, (tx + ux) / 2, (ty + uy) / 2);
	sx = tx;
	sy = ty;
	tx = ux;
	ty = uy;
	}
    ljLine(lx, ly, tx, ty);
}

#define FLAT 1
flatten(x2, y2, x3, y3, x4, y4)
int x2, y2, x3, y3, x4, y4; {

    int x1 = lx, y1 = ly, dx, dy, n1, n2, n;

    dx = x4 - x1;
    dy = y4 - y1;

    n1 = dy*(x2 - x1) - dx*(y2 - y1);
    n2 = dy*(x3 - x1) - dx*(y3 - y1);

    if (n1 < 0)
	n1 = -n1;
    if (n2 < 0)
	n2 = -n2;

    n = n1 > n2 ? n1 : n2;

    if (n*n / (dy*dy + dx*dx) <= FLAT*FLAT) {
	ljLine(lx, ly, x4, y4);
	lx = x4;
	ly = y4;
    } else {
	flatten( (x1 + x2)/2, (y1 + y2)/2,
		  (x1 + x2*2 + x3)/4, (y1 + y2*2 + y3)/4,
		  (x1 +3*x2 + 3*x3 + x4)/8, (y1 +3*y2 + 3*y3 + y4)/8);
	flatten( (x2 + x3*2 + x4)/4, (y2 + y3*2 + y4)/4,
		  (x3 + x4)/2, (y3 + y4)/2, x4, y4);
    }
}

static ljPos(x, y)
register int x, y; {
    printf(XY, x, y);
    lastYPos = y;
}


static
ljVerLine(x0, y0, y1)
register int x0, y0, y1; {
    if (y0 > y1) {
	register int t;
	t = y0;
	y0 = y1;
	y1 = t;
    }
    ljPos(x0-curthick, y0);
    printf("\033*c%da%db0P", curthick<<1, y1-y0);
    ljCircle (x0, y0, curthick, drawDot, 0);
    ljCircle (x0, y1, curthick, drawDot, 0);
}

static
ljHorLine(x0, y0, x1)
register int x0, y0, x1; {
    if (x0 > x1) {
	register int t;
	t = x0;
	x0 = x1;
	x1 = t;
    }
    ljPos(x0, y0-curthick);
    printf("\033*c%da%db0P", x1-x0, curthick<<1);
    ljCircle (x0, y0, curthick, drawDot, 0);
    ljCircle (x1, y0, curthick, drawDot, 0);
}

static
ljLine(x0, y0, x1, y1)
register int x0, y0, x1, y1; {
    DBP((D_PIC, "ljLine: (%d,%d)..(%d,%d), curthick: %d\n", x0, y0, x1, y1, curthick));
    if (y0 == y1)
	ljHorLine(x0, y0, x1);
    else if (x0 == x1)
	ljVerLine(x0, y0, y1);
    else
	ljHLine(x0, y0, x1, y1);

}

#define	abs(x)	((x) < 0? -(x): (x))

/*	Breshenham's line drawing algorithm.
	Extended to full 4 quadrant line drawing.
	Uses only adds/subtracts and <<1.
	Should run like the proverbial bat */

ljHLine(x0, y0, x1, y1)
int x0, y0, x1, y1; {
    extern int rev;
    if (abs(y1 - y0) > abs(x1 - x0)) {
	rev=1;
	lineseg2(y0, x0, y1, x1);
    } else {
	rev=0;
	lineseg2(x0, y0, x1, y1);
    }
    rev = 0;
}

int rev;
		
/* extend base algorithm to -45..45 & 135..225 */
lineseg2(x0, y0, x1, y1)
int x0, y0, x1, y1; {
    if (x1 < x0)
	lineseg(x1, y1, x0, y0);
    else
	lineseg(x0, y0, x1, y1);
}


static
lineseg (x0, y0, x1, y1)
int x0, y0, x1, y1; {
    register int dx, dy, iE, iNE, d, x, y, ddy;

    /* extend algorithm to -45..45 */
    dx = x1 - x0;
    dy = y1 - y0;
    if (dy < 0) {
	ddy = -1;
	dy = -dy;
    } else
	ddy = 1;

    /* Basic algorithm, 0..45 degrees */
    d = 2 * dy - dx;
    iE = 2 * dy;
    iNE = 2 * (dy - dx);
    x = x0;
    y = y0;
    linePoints(x, y);
    while (x < x1) {
	if (d <= 0) {
	    d = d + iE;
	    x = x + 1;
	} else {
	    d = d + iNE;
	    x = x + 1;
	    y = y + ddy;
	}
	linePoints(x, y);
    }
}

/* This procedure uses second-order partial differences to compute increments
   in the decision variable.  Assumes center of circle is at origin.
   It actually calculates only one octant, then circlePoints maps
   it into the rest of a circle */

ljCircle (origx, origy, radius, dotfcn, fill)
int origx, origy, radius, fill;
FUNC dotfcn;
{
    register int x, y, d, deltaE, deltaSE;

    DBP((dotfcn == drawDot ? D_PIC: 0, "ljCircle: (%d,%d),r%d,f%d\n",
	origx, origy, radius, fill));

    x = 0;
    y = radius;
    d = 1 - radius;
    deltaE = 3;
    deltaSE = -2 * radius + 5;
    circlePoints(origx, origy, x, y, dotfcn, fill);

    while (y > x) {
	if (d < 0) { /* select E */
	    d += deltaE;
	    deltaE += 2;
	    deltaSE += 2;
	    x++;
	} else {	/* select SE */
	    d += deltaSE;
	    deltaE += 2;
	    deltaSE += 4;
	    x++;
	    y--;
	}
	circlePoints(origx, origy, x, y, dotfcn, fill);
    }
}

circlePoints(ox, oy, x, y, dotfcn, fill)
int ox, oy, x, y, fill;
FUNC dotfcn; {
    if (fill != -1) {
	dofill(ox, x, oy+y, fill);
	dofill(ox, x, oy-y, fill);
	dofill(ox, y, oy+x, fill);
	dofill(ox, y, oy-x, fill);
    } else {
	dotfcn(ox+x, oy+y, dotfcn);
	dotfcn(ox+x, oy-y, dotfcn);
	dotfcn(ox-x, oy+y, dotfcn);
	dotfcn(ox-x, oy-y, dotfcn);
	dotfcn(ox+y, oy+x, dotfcn);
	dotfcn(ox+y, oy-x, dotfcn);
	dotfcn(ox-y, oy+x, dotfcn);
	dotfcn(ox-y, oy-x, dotfcn);
    }
}

/*	Describes octants, counterclockwise from 0.
	0: off
	1: on
	2: calculate
 */
static int	arcOctant[8], invert;
static double arcStart, arcEnd;

/* roughly the same as ljCircle, except octant selector gunge */
ljArc (x0, y0, xc, yc, x1, y1)
int x0, y0, xc, yc, x1, y1;
{
    register int x, y, d, deltaE, deltaSE, radius;
    double t, c;

    DBP((D_PIC, "ljArc: %d, %d, %d, %d, %d, %d\n",
	x0, y0, xc, yc, x1, y1));
    {
	int inttmp;

	inttmp = x0 + xc;
	x0 = xc;
	xc = inttmp;

	inttmp = y0 + yc;
	y0 = yc;
	yc = inttmp;
    }

    radius = sqrt((double) sqr(x0) + (double) sqr(y0)) + .5;
    arcStart = atan2((double) y0, (double) -x0);
    arcEnd = atan2((double) -y1, (double) x1);

    if (arcStart < 0)
	arcStart += RAD_C;
    if (arcEnd < 0)
	arcEnd += RAD_C;

    if (arcStart > arcEnd) {
	invert = 1;
	t = arcStart;
	arcStart = arcEnd;
	arcEnd = t;
    } else {
	invert = 0;
    }

    DBP((D_PIC, "ljArc: (%d,%d),r%d,inv:%d, (%f,%f)\n",
	xc, yc, radius, invert, arcStart, arcEnd));

    for (x = 0; x < 8; x++)
	arcOctant[x] = invert;

    x = 8 * arcStart / (2*M_PI);
    y = 8 * arcEnd / (2*M_PI);

    arcOctant[x] = 2;
    arcOctant[y] = 2;
    for (d = x+1; d < y; d++)
	arcOctant[d] = !invert;

    t = 2*M_PI / 8;
    for (x = 0; x < 8; x++) {
	c = x*t;
	DBP((D_PIC, "%d: %f:%f (%f:%f) %d (%d)\n", x, c, c+t, arcStart, arcEnd,
	    arcOctant[x], invert));
    }

    x = 0;
    y = radius;
    d = 1 - radius;
    deltaE = 3;
    deltaSE = -2 * radius + 5;
    arcPoints(xc, yc, x, y);

    while (y > x) {
	if (d < 0) { /* select E */
	    d += deltaE;
	    deltaE += 2;
	    deltaSE += 2;
	    x++;
	} else {	/* select SE */
	    d += deltaSE;
	    deltaE += 2;
	    deltaSE += 4;
	    x++;
	    y--;
	}
	arcPoints(xc, yc, x, y);
    }
}

/* if (octant on) ||
      (endpoints in this octant && calculation shows point should be here)
   put a dot
 */

#define arcDot(q, a, b, c, d) if (arcOctant[q] == 1 || \
    (arcOctant[q] == 2 && \
	((t = atan2((double) -b,(double) a)),t = (t<0?t+RAD_C:t),\
	invert ^ (t >= arcStart && t <= arcEnd)))) \
    drawDot(c,d)


arcPoints(ox, oy, x, y)
int ox, oy, x, y; {
    double t;

    arcDot(0, y, -x, ox+y, oy-x);
    arcDot(1, x, -y, ox+x, oy-y);
    arcDot(2, -x, -y, ox-x, oy-y);
    arcDot(3, -y, -x, ox-y, oy-x);
    arcDot(4, -y, x, ox-y, oy+x);
    arcDot(5, -x, y, ox-x, oy+y);
    arcDot(6, x, y, ox+x, oy+y);
    arcDot(7, y, x, ox+y, oy+x);
}

/*	the fill range 0..1000 selects one of these 8
	rows.  Each row describes a 8x8 fill pattern.
 */
static unsigned char fillpat[9][8] = {
    { 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff },
    { 0xc3, 0xa5, 0x24, 0xe9, 0xe9, 0x24, 0xa5, 0xc3 },
    { 0xe7, 0xc3, 0x24, 0x3c, 0x3c, 0x24, 0xc3, 0xe7 },
    { 0xc3, 0x81, 0x00, 0x18, 0x18, 0x00, 0x81, 0xc3 },
    { 0x81, 0x42, 0x24, 0x18, 0x18, 0x24, 0x42, 0x81 },
    { 0x81, 0x00, 0x00, 0x18, 0x18, 0x00, 0x00, 0x81 },
    { 0x00, 0x42, 0x00, 0x18, 0x18, 0x00, 0x42, 0x00 },
    { 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
    { 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 }
};

/* fill from (x-dx,y) to (x+dx, y) with the fill pattern.

    fill pattern is aligned on byte boundaries in raster,
    so theoretically, we could convert this to 0..7
    of bit-dinks, a multi-byte memset, and 0..7 more
    bit-dinks.  Profiling doesn't seem to show this as significant.
 */

dofill(x, dx, y, fill)
register int x, dx, y, fill; {
    register int t, td;
    static int lastind = -1;
    register int ind = fill/125;
    register unsigned char fp = fillpat[ind][y&7];

#ifdef DEBUG
    if (lastind != ind) {
	DBP((D_PIC, "dofill: (%d+-%d,%d) %d->%d, fp:%02x\n", x, dx, y, fill, ind, fp));
	lastind = ind;
    }
#endif

    asetbit(x, y);	/* make sure row is allocated */

    td = x + dx;

    for (t = x - dx; t < td; t++)
	if ((0x80 >> (t&7)) & fp)
	    setbit(t, y);
	else
	    clearbit(t, y);
}

EllipsePoints (ox, oy, x, y, fill)
int ox, oy, x, y, fill; {
    if (fill != -1) {
	dofill(ox, x, oy+y, fill);
	dofill(ox, x, oy-y, fill);
    } else {
	drawDot(ox + x, oy + y);
	drawDot(ox + x, oy - y);
	drawDot(ox - x, oy + y);
	drawDot(ox - x, oy - y);
    }
}


ljEllipse(ox, oy, a, b, fill)
int ox, oy, a, b, fill; {
    int x, y;
    double d1, d2;
    int sqra = a*a,
	sqrb = b*b;
    DBP((D_PIC, "ljEllipse: (%d,%d),xr%d,yr%d,f%d\n",
	ox, oy, a, b, fill));

    x = 0;
    y = b;
    d1 = sqrb - sqra * b + (double) sqra / 4;
    EllipsePoints(ox, oy, x, y, fill);

    while (sqra * (y - .5) > sqrb * (x + 1)) {
	if (d1 < 0) {
	    d1 = d1 + sqrb * (2 * x + 3);
	    x++;
	} else {
	    d1 = d1 + sqrb * (2 * x + 3) + sqra * (-2 * y + 2);
	    x++;
	    y--;
	}
	EllipsePoints(ox, oy, x, y, fill);
    }

    d2 = sqrb * sqr(x + .5) + sqra * sqr(y - 1) - sqra * sqrb;
    while (y > 0) {
	if (d2 < 0) {
	    d2 = d2 + sqrb * (2 * x + 2) + sqra * (-2 * y + 3);
	    x++;
	    y--;
	} else {
	    d2 = d2 + sqra * (-2 * y + 3);
	    y--;
	}
	EllipsePoints(ox, oy, x, y, fill);
    }
}

linePoints(x, y)
int x, y; {
    if (rev)
	drawDot(y,x);
    else
	drawDot(x,y);
}

/* General point drawing routine.  The idea is to draw a filled circle
   at each pixel with radius 1/2 the desired thickness.  Doing a filled
   circle is a lot of bit-dinking, so we optimize by drawing unfilled
   circles with minimum line-width, and rely on adjacent circles to fill
   in the hole.  Means that lines shorter than the line thickness
   may have some hollow bits.
 */
drawDot(x, y)
int x, y; {
    if (curthick <= 1) {
	asetbit(x, y);
	asetbit(x+1, y);
	asetbit(x, y+1);
	asetbit(x+1, y+1);
    } else
	ljCircle(x, y, curthick, asetbit, -1);
}

/*	RASTER ENGINE */

char **ljraster = NULL;
int thispage = 0;

/* only supports drawing resolution of 300 at the moment. */
#define DRAWRES	300
#define RESMULT (LJRESOLUTION / DRAWRES)
#define ESCLEN 25


int
nullscan(start, end)
register char *start, *end; {
    register char *cur;
    for (cur = start; cur < end && !*cur; cur++);
    return(cur - start);
}

int
pixblock(start, end)
register char *start, *end; {
    register char *cur;
    register int numnulls;
    if (end - start <= ESCLEN * 2)	/* no point optimizing */
	return(end - start);
    cur = start;

    while(cur < end) {

	for(; *cur && cur < end; cur++);

	numnulls = nullscan(cur, end);

	if (numnulls > ESCLEN)
	    return(cur - start);
	else
	    cur += numnulls;
    }
    return(end - start);
}

emitrast()
{
    int bperline;
    register char *ip, *lineanch, *nip;
    register currow, bcount;
    int linecount = 0;

    bperline = ((numx + 7) / 8);

    if (!thispage)
	return;

    printf("\033*t%dR", DRAWRES);

    for(currow = 0; currow < numy; currow++) {
	linecount++;
	if (!ljraster[currow])
	    continue;
	lineanch = ip = ljraster[currow];
	nip = ip + bperline;
	while (ip < nip && !*ip) ip++;
	if (ip >= nip)
	    continue;	/* line has no pixels */
	while (!*(nip - 1)) nip--;	/* truncate trailing nulls */
	while (ip < nip) {	/* inv: !*ip && !*nip */
	    bcount = pixblock(ip, nip);
	    printf("\033*p%dx%dY\033*r1A\033*b%dW",
		(ip - lineanch) * 8 * RESMULT, currow * RESMULT, bcount);
	    rastcount += bcount;
	    total += (bcount + ESCLEN);
	    fwrite(ip, 1, bcount, stdout);
	    fputs("\033*rB", stdout);
	    for(ip += bcount; ip < nip && !*ip; ip++);
	}
	free(ljraster[currow]);
	ljraster[currow] = (char *) NULL;
    }
    DBP((D_PIC, "emit raster: %d lines, %d rastbytes, total %d bytes\n",
	linecount, rastcount, total));
    rastcount = total = 0;
    thispage = 0;
}

asetbit(x, y)
register int x, y; {
    thispage = 1;
    if (y < 0 || y >= numy) {
	/*fprintf(stderr, "y (%d) too big or small (0..%d)\n", y, numy);*/
	return;
    }
    if (x < 0 || x >= numx) {
	/*fprintf(stderr, "x (%d) too big or small (0..%d)\n", x, numx);*/
	return;
    }
    if (!ljraster) {
	ljraster = (char **) mustmalloc(numy * sizeof(char *),
	    "ljraster pointers");
	clrarray(ljraster, numy * sizeof(char *));
    }
    if (!ljraster[y]) {
	ljraster[y] = (char *) mustmalloc(numx/8, "ljraster line");
	clrarray(ljraster[y], numx/8);
    }
    ljraster[y][x >> 3] |= (0x80 >> (x & 0x7));
}

setbit(x, y)
register int x, y; {
    if (!ljraster[y])
	return;
    ljraster[y][x >> 3] |= (0x80 >> (x & 0x7));
}

clearbit(x, y)
register int x, y; {
    if (!ljraster[y])
	return;
    ljraster[y][x >> 3] &= ~(0x80 >> (x & 0x7));
}


#endif
